/* HIN (Http INterceptor)

   Copyright 2013, 2014 by John K. Hinsdale.  See LICENSE.html for terms of use.
*/

/////////////////////////////////////////   Global vars

// Flag - the extension loaded w/ options applied?
var GLOBAL__loaded = false;

/* Hash of the request IDs that are matched.  This will allow us to
conduct the expensive match operation only once, then simply look here
for something to do with the request.  We'll also map the ID to the
actions that need to be taken per the config. */

var GLOBAL__matched_requests = {};

// Working copy of options
var GLOBAL__options = GLOBAL__options_default;

// Debug events: map request ID to captured events, track FIFO of requests, dirty flag
var GLOBAL__debug = { request: {}, fifo: [], dirty: false };

/////////////////////////////////////////   Functions


/*
 * onBeforeRequest
 *
 * NOT CALLED(!) on:
 *    https://clients2.google.com/availability/?s=gmail&a=viewinbox&c=scs&tm=1415390396312&zx=jylx2d70t3j0
 *    or more generally:
 *    https://clients2.google.com/availability
 * nor is it called on URLs that are configured to be loaded on startup, apparently.
 *
 * Setting startup page to "https://www.google.com/ triggers un-trapped calls to:
 *    POST https://accounts.google.com/ListAccounts?json=standard
 *
 */
function wr_handle_onBeforeRequest(info) {

    // Do nothing if globally disabled
    if ( !enabled() )
        return;
    // Never intercept a "chrome-extension://" URL so we don't
    // interfere with getting our own pages, icons, etc.
    if ( starts_with(info.url, "chrome-extension://") )
        return;

    // Result -- only thing we will set here is "cancel" and "redirectURL"
    var result = {};

    // Scan enabled rules and see which match the request
    var matched_rules = [];
    for (var r=0; r < GLOBAL__options.rules.length; r++) {
        var rule = GLOBAL__options.rules[r];
        // console.log("\n\n************** RULE:\n" + JSON.stringify(rule, null, 2));
        if ( rule.enabled && ! rule.compiled_match ) {
            // console.log("Rule '" + rule.name + "' has no compiled match");
            continue;
        }
        
        if ( rule.enabled && match_rule(rule, info) )
            matched_rules.push(rule);
    }

    // Do nothing if no matches
    if ( matched_rules.length == 0 )
        return result;

    // Map request ID to matched rules for use in later handlers.
    GLOBAL__matched_requests[info.requestId] = { matched_rules: matched_rules, debug: { enabled: false } };
    
    // Set up entry in debug capture.  It's important to do this in the entry event handler
    // in a synchronous way, so that asynchronous events which arrive at unpredictable times,
    // can be captured within the debug record for the request under it's ID.
    debug_event_start(info, matched_rules);

    // Execute actions for matched rules that need to run in this handler
    var event_name = "onBeforeRequest";
    for (var i=0; i < matched_rules.length; i++) {
        var rule = matched_rules[i];
        if ( event_name in rule.event_actions ) {
            var actions = rule.event_actions[event_name];
            for (var a=0; a < actions.length; a++) {
                var pair = actions[a];
                var actname = pair[0];
                var action = pair[1];
                // Handle cancel
                if ( action[0] == "cancel" ) {
                    debug_event_capture(info, event_name, rule.name, actname, "Cancel", null, null);
                    result.cancel = true;
                }
                // Handle redirect
                else if ( action[0] == "redirect" ) {
                    var url = substitute_url(action[1], info);
                    debug_event_capture(info, event_name, rule.name, actname, "Redirect to", url, null);
                    result.redirectUrl = url;
                }
                // Handle alert
                else if ( action[0] == "alert" ) {
                    var msg = substitute_url(action[1], info);
                    debug_event_capture(info, event_name, rule.name, actname, "Alert message", msg, null);
                    alert(msg);
                }
                // Handle debug (debug trace remainder of request/response)
                else if ( action[0] == "debug" ) {
                    debug_event_capture(info, event_name, rule.name, actname, "Debug", info.method, info.url);
                    GLOBAL__matched_requests[info.requestId].debug = { enabled: true, rule: rule.name, action: actname };
                }
            }
        }
    }

    // return a "BlockingResponse".  For onBeforeRequest event,
    // these optional params apply:
    //     cancel         boolean     Whether to drop the request
    //     redirectURL    string      Where to redirect
    return result;
}

/*
 * onBeforeSendHeaders
 */
function wr_handle_onBeforeSendHeaders(info) {
    // Do nothing if globally disabled
    if ( !enabled() )
        return;
    result = {};

    // Get and possibly adjust intended request headers.  Note that we will only get the
    // "requestHeaders" in the info if we ask for it in the "extraInfoSpec" below.
    // This is not documented(!)
    var m_result = modify_headers(info, "onBeforeSendHeaders", "request", info.requestHeaders);
    if ( ! m_result.did_anything )
        return result;

    // return a "BlockingResponse".  For onBeforeSendHeaders event,
    // these optional params apply:
    //     cancel         boolean       Whether to drop the request
    //     requestHeaders HttpHeaders   Request headers to send
    result.requestHeaders = m_result.headers;
    return result;
}

/*
 * onSendHeaders 
 */
function wr_handle_onSendHeaders(info) {
    // Do nothing if globally disabled
    if ( !enabled() )
        return;
    debug_actual_headers(info, "onSendHeaders", "request", info.requestHeaders);
    return {};
}

/*
 * onHeadersReceived
 */
function wr_handle_onHeadersReceived(info) {
    // Do nothing if globally disabled
    if ( !enabled() )
        return;
    result = {};
    
    // Get and possibly adjust intended response headers.  Note that we will only get the
    // "responseHeaders" in the info if we ask for it in the "extraInfoSpec" below.
    // This is not documented(!)
    var m_result = modify_headers(info, "onHeadersReceived", "response", info.responseHeaders);
    if ( ! m_result.did_anything )
        return result;

    // return a "BlockingResponse".  For onBeforeSendHeaders event,
    // these optional params apply:
    //     redirectURL    string        Where to redirect
    //     requestHeaders HttpHeaders   Response headers to pass through
    result.responseHeaders = m_result.headers;
    return result;
}

/*
 * onResponseStarted
 */
function wr_handle_onResponseStarted(info) {
    // Do nothing if globally disabled
    if ( !enabled() )
        return;
    debug_actual_headers(info, "onResponseStarted", "response", info.responseHeaders);
    return {};
}

/*
 * onAuthRequired
 */
function wr_handle_onAuthRequired(info) {
    // Do nothing if globally disabled
    if ( !enabled() )
        return;

    var result = {};
    
    // See if working on a request that matched rule(s)
    if ( ! (info.requestId in GLOBAL__matched_requests) )
        return result;
    var matched_rules = GLOBAL__matched_requests[info.requestId].matched_rules;
    var debug_info = GLOBAL__matched_requests[info.requestId].debug;
    var debug = debug_enabled() && debug_info.enabled;
    var event_name = "onAuthRequired";

    // Log the auth info requested
    if ( debug )
        debug_event_capture(info, event_name, debug_info.rule, debug_info.action, "Auth required",
                            "Scheme=[" + info.scheme + "] Realm=[" + info.realm + "]",
                            info.challenger.host + ":" + info.challenger.port);

    // Execute actions for matched rules that need to run in this handler
    for (var i=0; i < matched_rules.length; i++) {
        var rule = matched_rules[i];
        if ( event_name in rule.event_actions ) {
            var actions = rule.event_actions[event_name];
            for (var a=0; a < actions.length; a++) {
                var pair = actions[a];
                var actname = pair[0];
                var action = pair[1];
                // Handle authorize
                if ( action[0] == "authorize" ) {
                    var username = action[1];
                    var passwd = action[2];
                    // Check we aren't getting called repeatedly for the same request
                    if ( GLOBAL__matched_requests[info.requestId].auth_attempted )
                        debug_event_capture(info, event_name, rule.name, actname, "FAILED AUTHORIZE", username, passwd);
                    else {
                        var op = "Authorize";
                        if ( info.relam )
                            op += " (" + info.realm + ")";
                        debug_event_capture(info, event_name, rule.name, actname, op, username, passwd);
                        result.authCredentials = { username: username, password: passwd };
                        // Mark this request as having done an authorization attempt
                        GLOBAL__matched_requests[info.requestId].auth_attempted = true;
                    }
                }
            }
        }
    }
    return result;
}

/*
 * onCompleted
 */
function wr_handle_onCompleted(info) {
    // Do nothing if globally disabled
    if ( !enabled() )
        return {};
    if ( info.requestId in GLOBAL__matched_requests ) {
        var debug_info = GLOBAL__matched_requests[info.requestId].debug;
        var debug = debug_enabled() && debug_info.enabled;
        if ( debug )
            debug_event_capture(info, "onCompleted", debug_info.rule, debug_info.action, "SUCCESS", info.statusLine, null);
        
        // Remove request ID from global hash
        delete GLOBAL__matched_requests[info.requestId];

        // Shift debug FIFO
        debug_event_complete();
    }
    return {};
}

/*
 * onErrorOccurred
 */
function wr_handle_onErrorOccurred(info) {
    // Do nothing if globally disabled
    if ( !enabled() )
        return {};
    if ( info.requestId in GLOBAL__matched_requests ) {
        var debug_info = GLOBAL__matched_requests[info.requestId].debug;
        var debug = debug_enabled() && debug_info.enabled;
        if ( debug )
            debug_event_capture(info, "onErrorOccurred", debug_info.rule, debug_info.action, "FAILED", info.error, null);
        
        // Remove request ID from global hash
        delete GLOBAL__matched_requests[info.requestId];

        // Shift debug FIFO
        debug_event_complete();
    }
    return {};
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Log actual headers send/received
function debug_actual_headers(info, event_name, header_type, headers) {
    // See if working on a request that matched rule(s)
    if ( ! (info.requestId in GLOBAL__matched_requests) )
        return;
    var matched_rules = GLOBAL__matched_requests[info.requestId].matched_rules;
    var debug_info = GLOBAL__matched_requests[info.requestId].debug;
    var debug = debug_enabled() && debug_info.enabled;
    if ( !debug )
        return;
    for ( var i=0; i < headers.length; i++ )
        debug_event_capture(info, event_name, debug_info.rule, debug_info.action, "Actual header (" + header_type + ")",
                   headers[i].name, headers[i].value);
}

// Apply rules to request or response headers.  Return hash
// with "did_anything" set in which case return modified headers
// in value of key "headers".  Input "header_type" is either
// "request" or "response"
function modify_headers(info, event_name, header_type, headers) {

    var result = { did_anything: false };

    // Flag: request headers (vs. response headers)
    var is_req = true;
    if ( header_type == "response" )
        is_req = false;
    
    // See if working on a request that matched rule(s)
    if ( ! (info.requestId in GLOBAL__matched_requests) )
        return result;
    var matched_rules = GLOBAL__matched_requests[info.requestId].matched_rules;
    var debug_info = GLOBAL__matched_requests[info.requestId].debug;
    var debug = debug_enabled() && debug_info.enabled;
    
    if ( ! headers || headers.length == 0 ) {
        if ( debug )
            debug_event_capture(info, event_name, debug_info.rule, debug_info.action, "Debug", "No headers", null);
        return result;
    }
    
    // Scan rules for headers to add, override, or drop
    // Execute actions for matched rules that need to run in this handler
    // Commands recognized here are:
    //      drop_request_header <header_name>
    //      set_request_header <header_name> <value> [<add_if_absent>=true] [<overwrite_existing>=true]
    //      drop_response_header <header_name>
    //      set_response_header <header_name> <value> [<add_if_absent>=true] [<overwrite_existing>=true]
    
    var headers_to_drop = {};
    var headers_to_set = {};
    var drop_op_name = "drop_" + header_type + "_header";
    var set_op_name = "set_" + header_type + "_header";

    for (var i=0; i < matched_rules.length; i++) {
        var rule = matched_rules[i];
        if ( event_name in rule.event_actions ) {
            var actions = rule.event_actions[event_name];
            for (var a=0; a < actions.length; a++) {
                var pair = actions[a];
                var actname = pair[0];
                var action = pair[1];
                // Handle drop
                if ( action[0] == drop_op_name ) {
                    var hname_lc = action[1].toLowerCase();
                    headers_to_drop[hname_lc] = { header_name: action[1], rule: rule.name, action: actname };
                }
                // Handle set
                else if ( action[0] == set_op_name ) {
                    var hname_lc = action[1].toLowerCase();
                    var value = action[2];
                    var add_if_absent = true;
                    var overwrite_existing = true;
                    if ( action.length > 3 )
                        add_if_absent = action[3];
                    if ( action.length > 4 )
                        overwrite_existing = action[4];
                    headers_to_set[hname_lc] = { header_name: action[1], value: value,
                                                 add_if_absent: add_if_absent, overwrite_existing: overwrite_existing,
                                                 rule: rule.name, action: actname };
                }
            }
        }
    }
    
    // Scan headers intended to be sent/received and add/drop/replace per rules
    var headers_seen = {};
    var i = 0;
    while ( i < headers.length ) {
        var drop_header = false;
        var hn = headers[i].name;
        var hn_lc = hn.toLowerCase();
        headers_seen[hn_lc] = true;
        
        // Log pre-adjustment headers
        if ( debug )
            debug_event_capture(info, event_name, debug_info.rule, debug_info.action, "Header (" + header_type + ")", headers[i].name, headers[i].value);

        // See if global cookies disabled
        if ( ! GLOBAL__options.cookies && ( (is_req && (hn_lc == "cookie")) || (!is_req && (hn_lc == "set-cookie")) ) ) {
            // Drop the cookie header per global flag
            drop_header = true;
            // Log cookie drop per global default
            debug_event_capture(info, event_name, "[global]", "[cookies disabled]", "Dropping " + header_type + " cookie", hn, headers[i].value);
            
            /*
             * TODO: Check what is the effect on the browser's decision to cache a response
             * containing "Vary: Cookie" when we had previously dropped the "Cookie:" header
             * from the request? Does the browser construct the cache key using the Cookie:
             * header as it existed before this callback, or after?  If the cache decision is
             * based on the request header before the callback, then we may need to drop the
             * "Cookie" portion of the response's "Vary: header to compensate, if in fact
             * catching and modifying Vary: in the response, in a way that affects the cache
             * decision, is even allowed.
             */
        }
        // See if header is to be dropped by matched rule(s) (supersedes conflicting add)
        else if ( hn_lc in headers_to_drop ) {
            var inf = headers_to_drop[hn_lc];
            debug_event_capture(info, event_name, inf.rule, inf.action, "Drop " + header_type + " header", hn, headers[i].value);
            drop_header = true;
        }
        // See if need to overwrite header
        else if ( hn_lc in headers_to_set ) {
            var inf = headers_to_set[hn_lc];
            if ( inf.overwrite_existing ) {
                var newval = new_header_value(headers[i].value, inf.value);
                // Show new value, or if it did not change at all, show that
                var newval_disp = newval;
                if ( newval == headers[i].value )
                    newval_disp = "[unchanged]";
                debug_event_capture(info, event_name, inf.rule, inf.action, "Replace " + header_type + " header (old)", hn, headers[i].value);
                debug_event_capture(info, event_name, inf.rule, inf.action, "Replace " + header_type + " header (new)", hn, newval_disp);
                headers[i].value = newval;
            }
        }
        
        // Drop or replace the header
        if ( drop_header )
            headers.splice(i, 1);
        else
            // Use untouched or replaced header
            i++;
    }
    
    // Add headers not seen and where we need to set if absent
    var to_add = [];
    for ( var hn_lc in headers_to_set ) {
        var inf = headers_to_set[hn_lc];
        if ( ! inf.add_if_absent || (hn_lc in headers_seen) )
            continue;
        debug_event_capture(info, event_name, inf.rule, inf.action, "Add " + header_type + " header", inf.header_name, inf.value);
        to_add.push({ name: inf.header_name, value: inf.value });
    }
    if ( to_add.length > 0 )
        headers = headers.concat(to_add);

    result.did_anything = true;
    result.headers = headers;
    return result;
}

// Apply current options to toobar icon, tooltip and badge
function apply_options() {
    /*
     * Set icon and tooltip
     */
    var title = "HIN (Http INterceptor)";
    var badge = "";
    if ( GLOBAL__options.enabled ) {
        chrome.browserAction.setIcon({path: "icons/Fighter_jet_active_icon_19.png"});
        title += "\nStatus: Enabled";
    }
    else {
        chrome.browserAction.setIcon({path: "icons/Fighter_jet_disabled_icon_19.png"});
        title += "\nStatus: Disabled";
    }
    // Cookie blocking?
    if ( GLOBAL__options.cookies ) {
        title += "\nCookies: Allowed";
    }
    else {
        // Show cookie block in badge, but only if the debug badge is
        // not there (it's more important)
        if ( GLOBAL__options.enabled && !GLOBAL__options.debug )
            badge = "- CK";
        title += "\nCookies: Blocked";
    }
    // Show that option will take effect only after extension enabled
    if ( !GLOBAL__options.enabled )
        title += " (once enabled)";
        
    // Debug on?
    if ( GLOBAL__options.debug ) {
        title += "\nDebug capture: On";
        if ( GLOBAL__options.enabled )
            badge = "DBG!";
    }
    else {
        title += "\nDebug capture: Off";
    }
    // Show that option will take effect only after extension enabled
    if ( !GLOBAL__options.enabled )
        title += " (once enabled)";

    // Set the icon as enabled/disabled, and the title (tooltip) and "badge"
    if ( GLOBAL__options.enable_icon )
        chrome.browserAction.enable();
    else {
        // Disable the icon and show no badge nor tooltip
        chrome.browserAction.disable();
        badge = "";
    }
    chrome.browserAction.setBadgeText({text: badge});
    chrome.browserAction.setTitle({title: title});
}

// Check debug status from options
function debug_enabled() {
    return GLOBAL__options.debug;
}
// Check enabled status from options
function enabled() {
    return GLOBAL__options.enabled;
}

// Save options to storage
function save_options_to_storage(opts, callback) {
    chrome.storage.local.set(opts, callback);
}

// Match request against compiled rule
function match_rule(rule, info) {
    // console.log("Matching rule + '" + rule.name + "' against request " + info.method + " " + info.url + "\nexpr=" + JSON.stringify(rule.compiled_match, null, 2));
    return match_op(info, rule.compiled_match);
}
// Match request info against match operation [ <op>, [<args>] ]
function match_op(info, op_args) {
    // Handle raw boolean
    if ( (typeof op_args) == "boolean")
        return op_args;
    var op = op_args[0];
    // Handle logical operators
    if ( op == "not" )
        return ! match_op(info, op_args[1]);
    if ( op == "or" ) {
        for (var i=1; i < op_args.length; i++)
            if ( match_op(info, op_args[i]) )
                return true;
        return false;
    }
    if ( op == "and" ) {
        for (var i=1; i < op_args.length; i++)
            if ( ! match_op(info, op_args[i]) )
                return false;
        return true;
    }
    // Handle string operation on request field
    var fldval = null;
    if ( op == "url" )
        fldval = info.url;
    else if ( op == "method" )
        fldval = info.method;
    else
        return false;
    var string_op = op_args[1];
    var targ_list = op_args[2];

    // console.log("match_op: op=" + string_op + " args=" + JSON.stringify(targ_list, null, 2));

    // Apply test to targets in list until one matches (implied "or")
    if ( string_op == "contains" )
        return match_list_contains(fldval, targ_list);
    else if ( string_op == "starts" )
        return match_list_starts(fldval, targ_list);
    else if ( string_op == "ends" )
        return match_list_ends(fldval, targ_list);
    else if ( string_op == "equals" )
        return match_list_equals(fldval, targ_list);
    else if ( string_op == "matches" )
        return match_list_matches(fldval, targ_list);
    return false;
}

// Start debug event capture
function debug_event_start(info, rules) {
    if ( !debug_enabled() )
        return;
    var reqid = info.requestId;
    var dbg = GLOBAL__debug;
    // Init entry in debug map.
    if ( ! (reqid in dbg.request) ) {
        dbg.request[reqid] = { method: info.method, url: info.url, rules: rules, events: [] }
        // Push ID onto start of request FIFO
        dbg.fifo.unshift(reqid);
    }
    dbg.dirty = true;
}

// Capture debug event.  It is important that debug_event_start have been called
// with the request info before this function is called.  Because we block on
// execution of the entry point handler onBeforeRequest() and debug_event_start()
// is called there, this is all OK.
function debug_event_capture(info, event, rule, action, operation, arg1, arg2) {
    if ( !debug_enabled() )
        return;

    // console.log(pad(info.requestId, 6) + pad(event, 21) + pad(rule, 30) + pad(action, 30) + pad(operation, 21) + " " + arg1 + " " + arg2);

    // Append event info
    var reqid = info.requestId;
    var dbg = GLOBAL__debug;
    var evinfo = {event: event, rule: rule, action: action, operation: operation, arg1: arg1, arg2: arg2};
    dbg.request[reqid].events.push(evinfo);
    dbg.dirty = true;
}

// Shift FIFO on any request complete
function debug_event_complete() {
    if ( !debug_enabled() )
        return;
    // Truncate FIFO at end and remove from request map if we have exceeded the
    // size given in options
    var dbg = GLOBAL__debug;
    var maxfifo = GLOBAL__options.debug_requests_count;
    while ( dbg.fifo.length > maxfifo ) {
        var id = dbg.fifo.pop();
        delete dbg.request[id];
        dbg.dirty = true;
    }
}

// Reset the debug events data
function debug_event_reset() {
    GLOBAL__debug = { request: {}, fifo: [] }
    GLOBAL__debug.dirty = true;
}

// Substitute string with request info.  Placeholders "__method__" and "__url__"
// will be substituted with the raw method and URL.  Uppercase placehoders __METHOD__
// and __URL__ will be substituted with encoded form suitable for use in URL params
function substitute_url(s, info) {
    var method = info.method;
    var req_url = info.url;
    s = replace_all(s, "__method__", method);
    s = replace_all(s, "__url__", req_url);
    s = replace_all(s, "__METHOD__", encodeURIComponent(method)); // Encoding not really needed here
    s = replace_all(s, "__URL__", encodeURIComponent(req_url));
    return s;
}

// Derive new header value from old, using static value, or applying substitutions
function new_header_value(oldval, newval) {
    if ( (typeof newval) == "string" )
        return newval;
    var result = oldval;
    for (var i=0; i <newval.length; i++) {
        var from = newval[i][0];
        var to = newval[i][1];
        var re_info = get_regex(from);
        if ( re_info.valid )
            result = result.replace(re_info.regex, to);
    }
    return result;
}


//////////////////////  On startup: install handlers, then load options and set UI based on them

chrome.webRequest.onBeforeRequest.addListener(wr_handle_onBeforeRequest, { urls: [ "<all_urls>" ] }, ["blocking"]);
chrome.webRequest.onBeforeSendHeaders.addListener(wr_handle_onBeforeSendHeaders, { urls: [ "<all_urls>" ] }, ["blocking", "requestHeaders"]);
chrome.webRequest.onSendHeaders.addListener(wr_handle_onSendHeaders, { urls: [ "<all_urls>" ] }, ["requestHeaders"]);
chrome.webRequest.onHeadersReceived.addListener(wr_handle_onHeadersReceived, { urls: [ "<all_urls>" ] }, ["blocking", "responseHeaders"]);
chrome.webRequest.onResponseStarted.addListener(wr_handle_onResponseStarted, { urls: [ "<all_urls>" ] }, ["responseHeaders"]);
chrome.webRequest.onAuthRequired.addListener(wr_handle_onAuthRequired, { urls: [ "<all_urls>" ] }, ["blocking"]);
chrome.webRequest.onCompleted.addListener(wr_handle_onCompleted, { urls: [ "<all_urls>" ] });
chrome.webRequest.onErrorOccurred.addListener(wr_handle_onErrorOccurred, { urls: [ "<all_urls>" ] });

/*
// Show stored version
chrome.storage.local.get(
    { options_version: "NOT SET" },
    function(stored_opts) {
        console.log("STORED VERSION: " + stored_opts.options_version);
    }
);
*/

// Load options from disk
chrome.storage.local.get(
    GLOBAL__options_default,
    function(dirty_opts) {
        // Just in case compiled info was stored persistently, clean it up
        // This is also a good place to convert data in deprecated form
        var opts = clean_options(dirty_opts);
        compile_options(opts);
        // console.log("*********************************************");
        // console.log("Setting GLOBAL__OPTIONS:\n" + JSON.stringify(opts, null, 2));
        // console.log("*********************************************");
        GLOBAL__options = opts;
        apply_options();
        // Any time we set the options we need to flush out cached pages that may
        // have been handled under old rules
        chrome.webRequest.handlerBehaviorChanged();
        GLOBAL__loaded = true;
    }
);

// $Id: hin.js,v 1.43 2015/01/08 23:57:21 hin Exp $
