/*
  popup.js  - Javascript to support popup.html

  Copyright 2013, 2014 by John K. Hinsdale.  See LICENSE.html for terms of use.
*/


// Open FAQ (or active currently open FAQ)
function open_faq() {
    var faq_url = chrome.extension.getURL("faq.html");
    open_one_of(faq_url);
}

// Open the options tab (or activate the currently open options tab)
function open_options() {
    var options_url = chrome.extension.getURL("options.html");
    open_one_of(options_url);
}

// Open the debug tab (or activate the currently open debug tab)
function open_debug() {
    var debug_url = chrome.extension.getURL("debug.html");
    open_one_of(debug_url);
}

// Open a URL, but only one of them (closing others that may be open)
// If the URL is already open in the current window, activate that tab
function open_one_of(url) {
    // First close any tab(s) open w/ the URL, but not in the current window,
    // to prevent confusion
    chrome.tabs.query(
        {url: url, currentWindow: false},
        function(tabs) {
            for (var i=0; i < tabs.length; i++)
                chrome.tabs.remove(tabs[i].id);
        }
    );
    // Now look for an open one in the current window, and create one if needed
    chrome.tabs.query(
        {url: url, currentWindow: true},
        function(tabs) {
            if ( tabs.length > 0 )
                chrome.tabs.update(tabs[0].id, {active: true});
            else
                chrome.tabs.create({url: url});
            window.close();
        }
    );
}

// Reload all open options and debug tabs, and then close the popup
function reload_all_options_tabs() {
    var urls = [ "options.html", "debug.html" ];
    for (var u=0; u < urls.length; u++) {
        var url = chrome.extension.getURL(urls[u]);
        chrome.tabs.query(
            {url: url},
            function(tabs) {
                for (var i=0; i < tabs.length; i++)
                    // chrome.tabs.remove(tabs[i].id);
                    chrome.tabs.reload(tabs[i].id);
                window.close();
            }
        );
    }
}

// Togglers
function toggle_enabled() {
    toggle_and_close("enabled", null);
}
function toggle_cookies() {
    toggle_and_close("cookies", null);
}
function toggle_debug() {
    toggle_and_close("debug", null);
}
// Toggle rule enabled and close window
function toggle_rule_or_action(ev) {
    // Find the rule or action, toggle it, and close
    var id = ev.currentTarget.id;
    if ( starts_with(id, "rule") )
        toggle_and_close("rule", parseInt(id.substring(5)));
    else if ( starts_with(id, "action") )
        toggle_and_close("action", parseInt(id.substring(7)));
    else
        window.close();
}

// Populate the status based on the current options
function populate() {
    // Get current options from background page
    var b = chrome.extension.getBackgroundPage();
    var opts = b.GLOBAL__options;

    // Enabled - show icon as marker
    enabled_icon = document.getElementById('enabled_icon');
    if ( opts.enabled )
        enabled_icon.src = "icons/Fighter_jet_active_icon_19.png";
    else
        enabled_icon.src = "icons/Fighter_jet_disabled_icon_19.png";
    enabled_prompt = document.getElementById('enabled_prompt');
    if ( opts.enabled )
        enabled_prompt.innerHTML = "HIN is enabled";
    else
        enabled_prompt.innerHTML = "HIN is disabled";

    // Cookies
    cookies_marker = document.getElementById('cookies_marker');
    cookies_marker.style.color = truth_color(opts.cookies);
    cookies_marker.innerHTML = truth_marker(opts.cookies);
    cookies_prompt = document.getElementById('cookies_prompt');
    if ( opts.cookies )
        cookies_prompt.innerHTML = "Cookies allowed";
    else
        cookies_prompt.innerHTML = "Cookies blocked";
    
    // Debug capture
    debug_marker = document.getElementById('debug_marker');
    debug_marker.style.color = truth_color(opts.debug);
    debug_marker.innerHTML = truth_marker(opts.debug);
    debug_prompt = document.getElementById('debug_prompt');
    if ( opts.debug )
        debug_prompt.innerHTML = "Debug capture on";
    else
        debug_prompt.innerHTML = "Debug capture off";

    // Gray out cookies and debug status when extension is disabled to
    // indicate the modes have no effect
    if ( ! opts.enabled ) {
        var dis = "disabled";
        cookies_marker.className = dis;
        cookies_prompt.className = dis;
        debug_marker.className = dis;
        debug_prompt.className = dis;
    }

    // Put in the menu items for the rules and actions and the handler for each
    var table = document.getElementById("menu");

    var infs = [ { is_rule: true, type: "rule", objs: opts.rules, prompt: "RULES" },
                { is_rule: false, type: "action", objs: opts.actions, prompt: "ACTIONS" }
              ];
    var node = table.tBodies[0];
    for ( idx=0; idx < infs.length; idx++ ) {
        var inf = infs[idx];
        var objs = inf.objs;
        // Show HR divider row and prompt
        var tr = document.createElement("tr");
        tr.innerHTML = "<td>&nbsp;</td><td><hr></td>";
        node.appendChild(tr);
        tr = document.createElement("tr");
        tr.innerHTML = "<td>&nbsp;</td><td nowrap><b>" + inf.prompt + "</b> (click to toggle):</td>";
        node.appendChild(tr);
        // Append rows for rules and actions
        for (var i=0; i < objs.length; i++) {
            var obj = objs[i];
            // Format menu item row
            var tr = document.createElement("tr");
            tr.id = inf.type + "_" + i;
            var tr_class = "mitem";
            if ( !opts.enabled )
                tr_class += " disabled";
            tr.className = tr_class;
                
            // Set click handler for rule/action row
            if ( tr.addEventListener )
                tr.addEventListener("click", toggle_rule_or_action, false);
            else
                tr.attachEvent('click', toggle_rule_or_action);
        
            // First cell is indicator as to whether enabled
            var td = document.createElement("td");
            td.innerHTML = truth_marker(obj.enabled);
            td.style.color = truth_color(obj.enabled);
            tr.appendChild(td);

            // Second cell is name, with tooltip/title
            td = document.createElement("td");
            td.innerHTML = obj.name;
            td.style.whiteSpace = "nowrap";
            // Show description as tooltip.  Also add to tooltip rule/action info:
            //   - For rule, show enabled actions and match expression.
            //   - For action, show action expression
            var title = obj.description;
            if ( inf.is_rule ) {
                title += "\nEnabled action(s): " + obj.compiled_action.join(", ");
                title += "\nMatch: " + JSON.stringify(obj.match);
            }
            else
                title += "\nAction: " + JSON.stringify(obj.action);
            td.title = title;
            tr.appendChild(td);

            // Add the row
            node.appendChild(tr);
        }
    }
}

// Return truth color
function truth_color(b) {
    if ( b )
        return "green";
    return "red";
}
// Return truth marker
function truth_marker(b) {
    if ( b )
        return "&#10003;";    // On (checkmark)
    return "&#10007";        // Off ("X")}
}
    
// Toggle option and close popup window.  If option is a rule or action enabled status,
// also supply the index into the relevant list.
function toggle_and_close(opt, idx) {
    var b = chrome.extension.getBackgroundPage();
    var opts = b.GLOBAL__options;

    // Toggle working options and toolbar icon, badge, tooltip
    if ( opt == "rule" || opt == "action" ) {
        // A rule or action' enabled status was toggled
        var objs;
        if ( opt == "rule" )
            opts.rules[idx].enabled = ! b.GLOBAL__options.rules[idx].enabled;
        else if ( opt == "action" )
            opts.actions[idx].enabled = ! opts.actions[idx].enabled;
        else
            // Should never get here
            alert("Internal error, '" + opt + "' is neither rule or action");
    }
    else
        // A flag was toggled
        opts[opt] = ! opts[opt];
    
    // Apply and save options (to reflect the effect of flags on the toolbar icon and badge)
    b.apply_options();
    // Also reload any open options editor tabs
    var cleaned_opts = clean_options(opts);
    b.save_options_to_storage(cleaned_opts, reload_all_options_tabs);
    // Recompile so any disable/enables of rules and actions is effective
    compile_options(opts);
}

// Open extensions page
function open_browser_extensions() {
    chrome.windows.create({url: "chrome://extensions/?id=" + chrome.runtime.id, focused: true});
    window.close();
}

// Reload the extension
function reload_extension() {
    chrome.runtime.reload();
    window.close();
}

// Uninstall the extension
function uninstall_extension() {
    var cb = function() {
        // "Check" the lastError to avoid warning in console log
        if ( chrome.runtime.lastError )
            ;
        window.close();
    }
    chrome.management.uninstallSelf({showConfirmDialog: true}, cb);
}

///////////////////////////////////  Init


// Install handlers

// Toggle hanlders
document.getElementById('enabled').addEventListener('click', toggle_enabled);
document.getElementById('cookies').addEventListener('click', toggle_cookies);
document.getElementById('debug').addEventListener('click', toggle_debug);

// Click handler to open options, extensions ...
document.getElementById('options_link').addEventListener('click', open_options);
document.getElementById('debug_link').addEventListener('click', open_debug);
document.getElementById('extensions_link').addEventListener('click', open_browser_extensions);
document.getElementById('reload_link').addEventListener('click', reload_extension);
document.getElementById('uninstall_link').addEventListener('click', uninstall_extension);
document.getElementById('faq_link').addEventListener('click', open_faq);

// Populate the fields
populate();

// $Id: popup.js,v 1.23 2014/12/23 12:38:42 hin Exp $
